#include "scene.h"
#include "ppc.h"
#include "linesegment.h"
#include "triangle.h"
#include "tmesh.h"
#include <math.h>

Scene *scene;

#include <iostream>
#include <fstream>

using namespace std;


Scene::Scene() {

  perSessionInitHW = false;

  impPPC = 0;
  impQuad = 0;
  text = 0;

  rstep = 5.0f;
  tstep = 5.0f;

  canvas = 0;
  canvasppc = 0;
  ims[0] = 
    ims[1] = 
    ims[2] = 0;
  imppcs[0] = 
    imppcs[1] = 
    imppcs[2] = 0;
  currImage = 0;

  gui = new GUI();
  gui->show();

  int u0 = 0;
  int v0 = 0;
  int w = 600;
  int h = 400;
  fb = new FrameBuffer(u0, v0, w, h);
  fb->label("SW Frame Buffer");
  fb->show();

  hwfb = new FrameBuffer(u0, v0+fb->h+60, w, h);
  hwfb->isHW = true;
  hwfb->label("HW Frame Buffer");
  hwfb->show();

  gui->uiw->position(fb->w+20, 40);

  float hfov = 55.0f;
  ppc = new PPC(hfov, w, h);


  tmesh = new TMesh();
  tmesh->Load("geometry/teapot1K.bin");
  tmesh->Scale(25.0f);
  tmesh->TranslateTo(V3(0.0f, 0.0f, -50.0f));
#if 0
  for (int i = 0; i < tmesh->vertsN; i++) {
    tmesh->colors[i] = V3(1.0f, 0.0f, 0.0f);
  }
  V3 pointLightSource = ppc->C + V3(0.0f, 150.0f, -200.0f);
  tmesh->Light(pointLightSource, ppc->C);
#endif

  tmesh2 = 0;
#if 0
  tmesh2 = new TMesh();
  tmesh2->Load("geometry/teapot1K.bin");
  tmesh2->Scale(25.0f);
  tmesh2->TranslateTo(V3(0.0f, 0.0f, -37.5f));
  for (int i = 0; i < tmesh2->vertsN; i++) {
    tmesh2->colors[i] = V3(0.0f, 0.0f, 1.0f);
  }
  tmesh2->Light(pointLightSource, ppc->C);
#endif
  ppc2 = 0;
#if 0
  tmesh3 = new TMesh(V3(0.0f, 0.0f, -25.0f), 30.0f, 30.0f);
  text = new FrameBuffer(0, 0, 256, 256);
  unsigned int cols[2] = {0xFFFFFFFF, 0xFF000000};
  text->SetChecker(16, cols);
  tmesh3->texture = text;
#endif
  tmesh3 = 0;
  LoadView();

  Render();

}

void Scene::PanoramaSetup() {

  fb->hide();
  ims[0] = new FrameBuffer();
  ims[0]->Load("images/IMG_5550_15.tif");
  ims[1] = new FrameBuffer();
  ims[1]->Load("images/IMG_5551_15.tif");
  ims[2] = new FrameBuffer();
  ims[2]->Load("images/IMG_5552_15.tif");

  imppcs[0] = new PPC(50.0f, ims[0]->w, ims[0]->h);
  imppcs[1] = new PPC(50.0f, ims[0]->w, ims[0]->h);
  imppcs[1]->Pan(-20.0f);
  imppcs[2] = new PPC(50.0f, ims[0]->w, ims[0]->h);
  imppcs[2]->Pan(-20.0f);

  currImage = 0;
  ppc = imppcs[currImage+1];

  canvas = new FrameBuffer(0, 0, 1024, 600);
  gui->uiw->position(canvas->w+20, 40);
  canvasppc = new PPC(120.0f, canvas->w, canvas->h);
  canvas->label("Canvas");
  canvas->show();
  Render();
  canvas->redraw();

}

void Scene::LoadConstructionImages() {

  ims[0] = new FrameBuffer();
  ims[0]->Load("images/IMG_5550_15.tif");
  ims[1] = new FrameBuffer();
  ims[1]->Load("images/IMG_5551_15.tif");
  ims[2] = new FrameBuffer();
  ims[2]->Load("images/IMG_5552_15.tif");

  ims[0]->show();
  ims[1]->show();
  ims[2]->show();

  imppcs[0] = new PPC(50.0f, ims[0]->w, ims[0]->h);
  imppcs[1] = new PPC(50.0f, ims[0]->w, ims[0]->h);
  imppcs[2] = new PPC(50.0f, ims[0]->w, ims[0]->h);
  ifstream ifs("mydbg/reg.txt");
  ifs 
    >> (*imppcs[0]) 
    >> (*imppcs[1]) 
    >> (*imppcs[2]) ;

}

void Scene::DBG() {

  tmesh->isReflector = true;
  UpdateImpostor(tmesh, tmesh2);
  Render();
  return;

  PanoramaSetup();
  return;

  LoadConstructionImages();
  BuildPanorama(ims, imppcs, 3);
  return;

  V3 v(10.0f, 0.0f, 0.0f);
  V3 n(0.0f, 1.0f, 0.0f);
  V3 r = n.Reflect(v);
  return;

  int w = fb->w;
  int h = fb->h;
  float hfov = 45.0f;

  ppc2 = new PPC(*ppc);

  float desiredFocalLength = 20.0f;
  fb->ClearBuffers();
  tmesh->RenderFilled(ppc2, fb);
  tmesh2 = new TMesh(fb, ppc2);
  tmesh3 = new TMesh(fb, ppc2, ppc2->FocalLength()/desiredFocalLength);
  for (int i = 0; i < tmesh->vertsN; i++) {
    tmesh->colors[i] = tmesh->colors[i] / 4.0f;
    tmesh->colors[i][0] = tmesh->colors[i][0] * 8.0f;
    tmesh->colors[i][0] = (tmesh->colors[i][0] > 1.0f) ? 1.0f : 
      tmesh->colors[i][0];
  }

}

void Scene::RenderCanvas() {

  canvas->Set(0xFF000000);
  for (int i = 0; i <= currImage+1; i++) {
    canvas->Set(canvasppc, ims[i], imppcs[i]);
  }
  bool frameOnly = true;
  imppcs[currImage+1]->Visualize(canvasppc, canvas, 10.0f, 0xFF0000FF, frameOnly);
  canvas->redraw();

}

void Scene::Render() {

  if (hwfb)
    hwfb->redraw();

  if (canvas) {
    RenderCanvas();
    return;
  }

  fb->ClearBuffers();

  if (tmesh)
    tmesh->RenderFilled(ppc, fb);
  if (tmesh2)
    tmesh2->RenderFilled(ppc, fb);
  if (tmesh3)
    tmesh3->RenderFilled(ppc, fb);

  if (impQuad) {
    impQuad->RenderWF(0xFF00AAFF, ppc, fb);
//    impQuad->RenderFilled(ppc, fb);
  }

  fb->redraw();


}

void Scene::LoadView() {

  if (ppc) {
    ifstream ifs("mydbg/view.txt");
    ifs >> *ppc;
    ifs.close();
  }

  Render();
  fb->redraw();

}

void Scene::SaveView() {

  if (ppc) {
    ofstream ofs("mydbg/view.txt");
    ofs << *ppc;
    ofs.close();
  }

}

void Scene::NextPair() {

  currImage = (currImage+1) % 2;
  ppc = imppcs[currImage+1];

}


void Scene::SavePPCs() {

  ofstream ofs("mydbg/reg.txt");
  for (int i = 0; i < 3; i++) {
    ofs << *(imppcs[i]) << endl;
  }
  ofs.close();

}


void Scene::BuildPanorama(FrameBuffer **cims, PPC **cimppcs, 
    int cimN) {

  int fw = 512;
  for (int fi = 0; fi < 6; fi++)
    faceppcs[fi] = new PPC(90.0f, fw, fw);

  faceppcs[1]->Pan(-90.0f);
  faceppcs[2]->Pan(-90.0f);
  faceppcs[2]->Pan(-90.0f);
  faceppcs[3]->Pan(+90.0f);
  faceppcs[4]->Tilt(90.0f);
  faceppcs[5]->Tilt(-90.0f);

  for (int fi = 0; fi < 6; fi++) {
    faces[fi] = new FrameBuffer(0, 0, fw, fw);
    char facename[100];
    sprintf_s(facename, "face_%d.tif", fi);
    faces[fi]->label(facename);
    for (int v = 0; v < fw; v++) {
      for (int u = 0; u < fw; u++) {
        V3 P = faceppcs[fi]->GetPoint(u, v, 1.0f);
        for (int ci = 0; ci < cimN; ci++) {
          V3 projP;
          if (!cimppcs[ci]->Project(P, projP))
            continue;
          if (
            projP[0] < 0.0f || projP[0] > (float) cims[ci]->w ||
            projP[1] < 0.0f || projP[1] > (float) cims[ci]->h)
            continue;
          faces[fi]->Set(u, v, cims[ci]->Get((int)projP[0], (int)projP[1]));
          break;
        }
      }
    }
    faces[fi]->show();
  }


}

void Scene::UpdateImpostor(TMesh *reflector, TMesh *diffuse) {

  // set camera to acquire impostor
  V3 bbox[2];
  diffuse->BoundingBox(bbox);
  float d = (bbox[1]-bbox[0]).Length();
  V3 Cr = reflector->GetCenter();
  V3 Cd = diffuse->GetCenter();
  float f = (Cd-Cr).Length();
  float fov = 2.0f*atan(d / 2.0f / f)/3.1415f*180.0f;
  int impW = 512;
  int impH = 512;
  impPPC = new PPC(fov, impW, impH);
  impPPC->PositionAndOrient(Cr, Cd, V3(0.0f, 1.0f, 0.0f));

  // acquire impostor
  FrameBuffer *impFB = new FrameBuffer(0, fb->h+60, impW, impH);
  impFB->ClearBuffers();
  diffuse->RenderFilled(impPPC, impFB);
  impFB->show();

  impQuad = new TMesh();
  impQuad->SetBillboard(impPPC, impFB, Cd);  

}

void Scene::RenderHW() {

  if (!perSessionInitHW)
    PerSessionInitHW();

  fb->ClearBuffersHW();

  ppc->SetViewHW();

  if (tmesh)
    tmesh->RenderFilledHW();

}

void Scene::PerSessionInitHW() {

  glEnable(GL_DEPTH_TEST);

  // initialize textures

}